



if not exist %LIBRARY_INC%\zfp.h exit 1
IF %ERRORLEVEL% NEQ 0 exit /B 1
if not exist %LIBRARY_BIN%\zfp.dll exit 1
IF %ERRORLEVEL% NEQ 0 exit /B 1
if not exist %LIBRARY_LIB%\zfp.lib exit 1
IF %ERRORLEVEL% NEQ 0 exit /B 1
if not exist %LIBRARY_PREFIX%\lib\cmake\zfp\zfp-config.cmake exit 1
IF %ERRORLEVEL% NEQ 0 exit /B 1
where zfp
IF %ERRORLEVEL% NEQ 0 exit /B 1
exit /B 0
